use crate::{models::User, prelude::*};
use futures::stream;
use itertools::Itertools;
use parking_lot::Mutex;
use std::sync::Arc;
use tokio::task::spawn_blocking;

/// Ecryptes the account and adds it to the database
/// If any of these steps fail, the account name will be added to the failed vector
async fn encrypt_account(
    account: DecryptedAccount,
    user_id: u64,
    db: &Pool,
    master_pass: Arc<str>,
    failed: &Mutex<&mut Vec<String>>,
) {
    let name = account.name.clone();
    match spawn_blocking(move || account.into_account(user_id, &master_pass)).await {
        Ok(account) => match account.insert(db).await {
            Ok(()) => (),
            Err(_) => failed.lock().push(name),
        },
        _ => failed.lock().push(name),
    }
}

/// Gets the master password, encryptes and adds the accounts to the DB
#[allow(clippy::too_many_arguments)]
async fn get_master_pass(
    bot: Throttle<Bot>,
    msg: Message,
    db: Pool,
    dialogue: MainDialogue,
    mut ids: MessageIds,
    locale: LocaleRef,
    user: User,
    master_pass: String,
) -> crate::Result<()> {
    dialogue.exit().await?;

    let user_id = msg.get_user_id()?;
    let mut failed = Vec::new();

    {
        let failed = Mutex::new(&mut failed);
        let master_pass: Arc<str> = master_pass.into();

        stream::iter(user.accounts)
            .for_each_concurrent(3, |account| {
                encrypt_account(account, user_id, &db, master_pass.clone(), &failed)
            })
            .await;
    }

    let text = if failed.is_empty() {
        locale.success.to_owned()
    } else {
        format!(
            "{}:\n{}",
            locale.couldnt_create_following_accounts,
            failed.into_iter().format("\n")
        )
    };
    ids.alter_message(&bot, text, deletion_markup(locale), None)
        .await?;
    Ok(())
}

handler!(get_user(user: User),send_master_password, State::GetMasterPass, get_master_pass);
first_handler!(import, send_json_file, State::GetUser, get_user);
