use crate::{
    change_state, cryptography::hashing::HashedBytes, entity::locale::LocaleType,
    locales::LocaleTypeExt, prelude::*,
};
use tokio::task::spawn_blocking;

#[allow(clippy::too_many_arguments)]
async fn get_master_pass2(
    bot: Throttle<Bot>,
    msg: Message,
    db: Pool,
    dialogue: MainDialogue,
    mut ids: MessageIds,
    locale: LocaleRef,
    hash: HashedBytes<[u8; 64], [u8; 64]>,
    master_pass: String,
) -> crate::Result<()> {
    dialogue.exit().await?;
    let from = msg.get_user()?;
    let user_id = msg.get_user_id()?;

    if !hash.verify(master_pass.as_bytes()) {
        ids.alter_message(
            &bot,
            &locale.master_password_dont_match,
            deletion_markup(locale),
            None,
        )
        .await?;

        return Ok(());
    }

    let model = MasterPass {
        user_id,
        password_hash: hash.hash.to_vec(),
        salt: hash.salt.to_vec(),
    };
    let locale_type = from
        .language_code
        .as_deref()
        .and_then(LocaleType::from_language_code)
        .unwrap_or_default();
    model.insert(&db, locale_type).await?;

    ids.alter_message(&bot, &locale.success, deletion_markup(locale), None)
        .await?;

    Ok(())
}

async fn get_master_pass(
    bot: Throttle<Bot>,
    _: Message,
    _: Pool,
    dialogue: MainDialogue,
    mut ids: MessageIds,
    locale: LocaleRef,
    master_pass: String,
) -> crate::Result<()> {
    let hash = spawn_blocking(move || HashedBytes::new(master_pass.as_bytes())).await?;

    ids.alter_message(&bot, &locale.send_master_password_again, None, None)
        .await?;

    change_state!(
        dialogue,
        ids,
        (hash),
        State::GetNewMasterPass,
        get_master_pass2
    );

    Ok(())
}

/// Handles /`set_master_pass` command
pub async fn set_master_pass(
    bot: Throttle<Bot>,
    msg: Message,
    dialogue: MainDialogue,
    db: Pool,
    locale: LocaleRef,
) -> crate::Result<()> {
    let user_id = msg.get_user_id()?;
    if MasterPass::exists(user_id, &db).await? {
        bot.send_message(msg.chat.id, &locale.master_password_is_set)
            .reply_markup(deletion_markup(locale))
            .await?;
        return Ok(());
    }
    let previous = bot
        .send_message(msg.chat.id, &locale.send_new_master_password)
        .await?;

    change_state!(
        dialogue,
        &previous,
        (),
        State::GetNewMasterPass,
        get_master_pass
    );

    Ok(())
}
