use super::Pool;
use futures::{Stream, TryStreamExt};
use sqlx::{Executor, FromRow, MySql, query, query_as};
use tokio::task::spawn_blocking;

#[derive(Clone, Debug, PartialEq, Eq, FromRow, Default)]
pub struct Account {
    pub user_id: u64,
    pub name: String,
    pub salt: Vec<u8>,
    pub enc_login: Vec<u8>,
    pub enc_password: Vec<u8>,
}

impl Account {
    // Inserts the account into DB
    pub async fn insert(&self, pool: &Pool) -> super::Result<()> {
        query!(
            "INSERT INTO account VALUES (?, ?, ?, ?, ?)",
            self.user_id,
            self.name,
            self.salt,
            self.enc_login,
            self.enc_password
        )
        .execute(pool)
        .await
        .map(|_| ())
    }

    /// Gets all user's account from DB
    pub fn get_all(user_id: u64, pool: &Pool) -> impl Stream<Item = super::Result<Self>> + '_ {
        query_as!(Self, "SELECT * FROM account WHERE user_id = ?", user_id).fetch(pool)
    }

    /// Streams the names of the user accounts
    pub fn get_names(user_id: u64, pool: &Pool) -> impl Stream<Item = super::Result<String>> + '_ {
        query!(
            "SELECT name FROM account WHERE user_id = ? ORDER BY name",
            user_id
        )
        .fetch(pool)
        .map_ok(|record| record.name)
    }

    /// Checks if the account exists
    pub async fn exists(user_id: u64, account_name: &str, pool: &Pool) -> super::Result<bool> {
        query_as::<_, (bool,)>(
            "SELECT EXISTS(SELECT * FROM account WHERE user_id = ? AND name = ? LIMIT 1) as value",
        )
        .bind(user_id)
        .bind(account_name)
        .fetch_one(pool)
        .await
        .map(|(exists,)| exists)
    }

    /// Gets the account from the DB
    pub async fn get(user_id: u64, account_name: &str, pool: &Pool) -> super::Result<Option<Self>> {
        query_as!(
            Self,
            "SELECT * FROM account WHERE user_id = ? AND name = ?",
            user_id,
            account_name
        )
        .fetch_optional(pool)
        .await
    }

    /// Searches for an account
    pub async fn search(
        user_id: u64,
        search_str: String,
        pool: &Pool,
    ) -> crate::Result<Vec<String>> {
        use fuzzy_matcher::{FuzzyMatcher, skim::SkimMatcherV2};

        let mut accounts: Vec<String> = Self::get_names(user_id, pool).try_collect().await?;

        spawn_blocking(move || {
            let matcher = SkimMatcherV2::default().ignore_case();
            accounts.retain(|name| matcher.fuzzy_match(name, &search_str).is_some());
            accounts
        })
        .await
        .map_err(Into::into)
    }

    // Deletes the account from DB
    pub async fn delete(user_id: u64, name: &str, pool: &Pool) -> super::Result<()> {
        query!(
            "DELETE FROM account WHERE user_id = ? AND name = ?",
            user_id,
            name
        )
        .execute(pool)
        .await
        .map(|_| ())
    }

    /// Deletes all the user's accounts from DB
    pub async fn delete_all(
        user_id: u64,
        pool: impl Executor<'_, Database = MySql>,
    ) -> super::Result<()> {
        query!("DELETE FROM account WHERE user_id = ?", user_id)
            .execute(pool)
            .await
            .map(|_| ())
    }

    /// Gets a name by a SHA256 hash of the name
    pub async fn get_name_by_hash(
        user_id: u64,
        hash: &[u8],
        pool: &Pool,
    ) -> super::Result<Option<String>> {
        let hash = hex::encode(hash);
        let name = query!(
            "SELECT `name` FROM `account` WHERE SHA2(`name`, 256) = ? AND `user_id` = ?;",
            hash,
            user_id,
        )
        .fetch_optional(pool)
        .await?;

        Ok(name.map(|record| record.name))
    }
    pub async fn get_salt(user_id: u64, name: &str, pool: &Pool) -> super::Result<Option<Vec<u8>>> {
        let salt = query!(
            "SELECT salt FROM account WHERE user_id = ? AND name = ?",
            user_id,
            name
        )
        .fetch_optional(pool)
        .await?;

        Ok(salt.map(|record| record.salt))
    }
    pub async fn update_name(
        user_id: u64,
        original_name: &str,
        new_name: &str,
        pool: &Pool,
    ) -> super::Result<bool> {
        query!(
            "UPDATE account SET name = ? WHERE user_id = ? AND name = ?",
            new_name,
            user_id,
            original_name
        )
        .execute(pool)
        .await
        .map(|result| result.rows_affected() != 0)
    }
    pub async fn update_login(
        user_id: u64,
        name: &str,
        login: Vec<u8>,
        pool: &Pool,
    ) -> super::Result<bool> {
        query!(
            "UPDATE account SET enc_login = ? WHERE user_id = ? AND name = ?",
            login,
            user_id,
            name
        )
        .execute(pool)
        .await
        .map(|result| result.rows_affected() != 0)
    }
    pub async fn update_password(
        user_id: u64,
        name: &str,
        password: Vec<u8>,
        pool: &Pool,
    ) -> super::Result<bool> {
        query!(
            "UPDATE account SET enc_password = ? WHERE user_id = ? AND name = ?",
            password,
            user_id,
            name
        )
        .execute(pool)
        .await
        .map(|result| result.rows_affected() != 0)
    }
}
