use crate::{errors::HandlerUsed, prelude::*};
use futures::future::BoxFuture;

/// A generic state handler. It checks for "/cancel" messages and runs the provided validation function
pub async fn generic<F>(
    bot: Throttle<Bot>,
    msg: Message,
    db: Pool,
    dialogue: MainDialogue,
    check: F,
    locale: LocaleRef,
    handler: PackagedHandler<String>,
) -> crate::Result<()>
where
    for<'a> F: FnOnce(
            &'a Message,
            &'a Pool,
            LocaleRef,
            &'a str,
        ) -> BoxFuture<'a, crate::Result<Option<String>>>
        + Send,
{
    let mut handler = handler.lock().await;
    if handler.func.is_none() {
        let _ = dialogue.exit().await;
        return Err(HandlerUsed.into());
    }

    let Some(text) = msg.text().map(str::trim) else {
        handler
            .previous
            .alter_message(&bot, &locale.couldnt_get_message_text, None, None)
            .await?;
        return Ok(());
    };

    if text == "/cancel" {
        dialogue.exit().await?;
        handler
            .previous
            .alter_message(
                &bot,
                &locale.successfully_canceled,
                deletion_markup(locale),
                None,
            )
            .await?;
        return Ok(());
    }

    if let Some(text) = check(&msg, &db, locale, text).await? {
        handler
            .previous
            .alter_message(&bot, text, None, None)
            .await?;
        return Ok(());
    }

    let previous = handler.previous;
    let func = handler.func.take().unwrap();
    drop(handler);
    let text = text.to_owned();

    if let Err(err) = func(bot, msg, db, dialogue.clone(), previous, locale, text).await {
        let _ = dialogue.exit().await;
        return Err(err);
    }
    Ok(())
}
